/*
 *  pv_tile_rotation_x11.c (for X11)
 *  ProteinViewer
 *
 *  Created by Naoto Morikawa on 6/23/06.
 *  Copyright 2006 Naoto Morikawa. All rights reserved.
 *
 *  2006-10-01 version 0.1 release
 *  2006-11-20 version 0.2 release
 */

#include <stdlib.h>
#include <GL/glut.h>
#include <stdio.h>
#include <math.h>
#include "pv_main.h"
#include "pv_read_data.h"
#include "pv_primitives.h"

extern struct pv_macro my_macro;
extern struct pe_protein my_protein;
extern struct pe_coor_system my_coor;
extern struct pe_trn_table my_trn_table[];

void print_mtrx(struct pe_matrix * pt_M)
{
	fprintf(STDOUT, "print_mat() :               | %8.3f, %8.3f, %8.3f |\n",
		pt_M->a11, pt_M->a12, pt_M->a13);
	fprintf(STDOUT, "print_mat() :         Mat = | %8.3f, %8.3f, %8.3f |\n",
		pt_M->a21, pt_M->a22, pt_M->a23);
	fprintf(STDOUT, "print_mat() :               | %8.3f, %8.3f, %8.3f |\n",
		pt_M->a31, pt_M->a32, pt_M->a33);
	return;
}

void print_vec(struct pe_vector * pt_vec)
{
	fprintf(STDOUT, "print_vec() :         vec = ( %8.3f, %8.3f, %8.3f )\n",
		pt_vec->x, pt_vec->y, pt_vec->z);
	return;
}

/* c = a x b */
void cross(struct pe_vector * pt_c, struct pe_vector * pt_a,
	   struct pe_vector * pt_b)
{
	pt_c->x = pt_a->y * pt_b->z - pt_a->z * pt_b->y;
	pt_c->y = pt_a->z * pt_b->x - pt_a->x * pt_b->z;
	pt_c->z = pt_a->x * pt_b->y - pt_a->y * pt_b->x;
	return;
}

float determinant(struct pe_matrix * pt_mat)
{
	float det;
	det = pt_mat->a11 * (pt_mat->a22 * pt_mat->a33
			     - pt_mat->a23 * pt_mat->a32)
		- pt_mat->a12 * (pt_mat->a21 * pt_mat->a33
				 - pt_mat->a23 * pt_mat->a31)
		+ pt_mat->a13 * (pt_mat->a21 * pt_mat->a32
				 - pt_mat->a22 * pt_mat->a31);
	return det;
}

void inverse_mat(struct pe_matrix * pt_inv, struct pe_matrix * pt_mat)
{
	float detA11, detA21, detA31;
	float detA12, detA22, detA32;
	float detA13, detA23, detA33;
	float detA;

	detA = determinant(pt_mat);
	if (!FLOAT_NEQ(detA)) {
		fprintf(STDERR, "inverse_mat() : ERROR  - |A| == %8.3f -\n",
			detA);
		exit(1);
	}
	/* The first row */
	detA11 = pt_mat->a22 * pt_mat->a33 - pt_mat->a23 * pt_mat->a32;
	pt_inv->a11 = detA11 / detA;
	detA21 = pt_mat->a12 * pt_mat->a33 - pt_mat->a13 * pt_mat->a32;
	pt_inv->a12 = -1 * detA21 / detA;
	detA31 = pt_mat->a12 * pt_mat->a23 - pt_mat->a13 * pt_mat->a22;
	pt_inv->a13 = detA31 / detA;
	/* The second row */
	detA12 = pt_mat->a21 * pt_mat->a33 - pt_mat->a23 * pt_mat->a31;
	pt_inv->a21 = -1 * detA12 / detA;
	detA22 = pt_mat->a11 * pt_mat->a33 - pt_mat->a13 * pt_mat->a31;
	pt_inv->a22 = detA22 / detA;
	detA32 = pt_mat->a11 * pt_mat->a23 - pt_mat->a13 * pt_mat->a21;
	pt_inv->a23 = -1 * detA32 / detA;
	/* The third row */
	detA13 = pt_mat->a21 * pt_mat->a32 - pt_mat->a22 * pt_mat->a31;
	pt_inv->a31 = detA13 / detA;
	detA23 = pt_mat->a11 * pt_mat->a32 - pt_mat->a12 * pt_mat->a31;
	pt_inv->a32 = -1 * detA23 / detA;
	detA33 = pt_mat->a11 * pt_mat->a22 - pt_mat->a12 * pt_mat->a21;
	pt_inv->a33 = detA33 / detA;
	return;
}

void mat_mul(struct pe_matrix * pt_C, struct pe_matrix * pt_A,
	     struct pe_matrix * pt_B)
{				/* C = AB */
	/* The first row */
	pt_C->a11 = pt_A->a11 * pt_B->a11 + pt_A->a12 * pt_B->a21
		+ pt_A->a13 * pt_B->a31;
	pt_C->a12 = pt_A->a11 * pt_B->a12 + pt_A->a12 * pt_B->a22
		+ pt_A->a13 * pt_B->a32;
	pt_C->a13 = pt_A->a11 * pt_B->a13 + pt_A->a12 * pt_B->a23
		+ pt_A->a13 * pt_B->a33;
	/* The second row */
	pt_C->a21 = pt_A->a21 * pt_B->a11 + pt_A->a22 * pt_B->a21
		+ pt_A->a23 * pt_B->a31;
	pt_C->a22 = pt_A->a21 * pt_B->a12 + pt_A->a22 * pt_B->a22
		+ pt_A->a23 * pt_B->a32;
	pt_C->a23 = pt_A->a21 * pt_B->a13 + pt_A->a22 * pt_B->a23
		+ pt_A->a23 * pt_B->a33;
	/* The third row */
	pt_C->a31 = pt_A->a31 * pt_B->a11 + pt_A->a32 * pt_B->a21
		+ pt_A->a33 * pt_B->a31;
	pt_C->a32 = pt_A->a31 * pt_B->a12 + pt_A->a32 * pt_B->a22
		+ pt_A->a33 * pt_B->a32;
	pt_C->a33 = pt_A->a31 * pt_B->a13 + pt_A->a32 * pt_B->a23
		+ pt_A->a33 * pt_B->a33;
	return;
}

int chk_eq(struct pe_matrix * pt_A, struct pe_matrix * pt_B)
{				/* return 1 if A == B, else return 0 */
	if (!FLOAT_NEQ(my_coor.ave_dis)) {
		fprintf(STDERR, "chk_eq() : ERROR  my_coor.ave_dis= %8.3f\n",
			my_coor.ave_dis);
		exit(1);
	}
	if (FLOAT_NEQ((pt_A->a11 - pt_B->a11) / my_coor.ave_dis))
		return FAIL;
	if (FLOAT_NEQ((pt_A->a12 - pt_B->a12) / my_coor.ave_dis))
		return FAIL;
	if (FLOAT_NEQ((pt_A->a13 - pt_B->a13) / my_coor.ave_dis))
		return FAIL;

	if (FLOAT_NEQ((pt_A->a21 - pt_B->a21) / my_coor.ave_dis))
		return FAIL;
	if (FLOAT_NEQ((pt_A->a22 - pt_B->a22) / my_coor.ave_dis))
		return FAIL;
	if (FLOAT_NEQ((pt_A->a23 - pt_B->a23) / my_coor.ave_dis))
		return FAIL;

	if (FLOAT_NEQ((pt_A->a31 - pt_B->a31) / my_coor.ave_dis))
		return FAIL;
	if (FLOAT_NEQ((pt_A->a32 - pt_B->a32) / my_coor.ave_dis))
		return FAIL;
	if (FLOAT_NEQ((pt_A->a33 - pt_B->a33) / my_coor.ave_dis))
		return FAIL;

	return SUCCESS;
}

void set_vector(struct pe_vector * pvec, float val_x, float val_y, float val_z)
{
	float wrk_length;
	
	wrk_length = sqrt(val_x * val_x + val_y * val_y + val_z * val_z);
	pvec->x = val_x / wrk_length;
	pvec->y = val_y / wrk_length;
	pvec->z = val_z / wrk_length;
	pvec->length = 1.0;
	return;
}

/* T(vec_a, vec_bb, vec_c) = (ex, ey, ez) */
/*     ==> T = inverse(A),	 */
/*           | ax bbx cx |	 */
/* where A = | ay bby cy |	 */
/*           | az bbz cz |	 */
void compute_matrix_T(struct pe_matrix * pT, struct pe_vector * pvec_a,
		      struct pe_vector * pvec_bb, struct pe_vector * pvec_c)
{
	float detA;
	
	detA = pvec_a->x * (pvec_bb->y * pvec_c->z - pvec_bb->z * pvec_c->y)
	- pvec_bb->x * (pvec_a->y * pvec_c->z - pvec_a->z * pvec_c->y)
	+ pvec_c->x * (pvec_a->y * pvec_bb->z - pvec_a->z * pvec_bb->y);
	if (0) fprintf(STDERR, "compute_matrix_T() : DEBUG  detA= %f\n", detA);
	if (FLOAT_NEQ((detA - 1.0) / my_coor.ave_dis)) {
		fprintf(STDERR, "compute_matrix_T() : ERROR  |T|= %f <==> 0 (ave_dis= %f)\n",
			detA, my_coor.ave_dis);
		exit(1);
	}
	/* The first row */
	pT->a11 = (pvec_bb->y * pvec_c->z - pvec_bb->z * pvec_c->y) / detA;
	pT->a12 = -1.0 * (pvec_bb->x * pvec_c->z - pvec_bb->z * pvec_c->x)
		/ detA;
	pT->a13 = (pvec_bb->x * pvec_c->y - pvec_bb->y * pvec_c->x) / detA;
	/* The second row */
	pT->a21 = -1.0 * (pvec_a->y * pvec_c->z - pvec_a->z * pvec_c->y) / detA;
	pT->a22 = (pvec_a->x * pvec_c->z - pvec_a->z * pvec_c->x) / detA;
	pT->a23 = -1.0 * (pvec_a->x * pvec_c->y - pvec_a->y * pvec_c->x) / detA;
	/* The third row */
	pT->a31 = (pvec_a->y * pvec_bb->z - pvec_a->z * pvec_bb->y) / detA;
	pT->a32 = -1.0 * (pvec_a->x * pvec_bb->z - pvec_a->z * pvec_bb->x) /
		detA;
	pT->a33 = (pvec_a->x * pvec_bb->y - pvec_a->y * pvec_bb->x) / detA;
	return;
}

/* R = rotation around z-ax */
/* |cosQ -sinQ  0|	 */
/* R= |sinQ  cosQ  0|	 */
/* |  0     0   1|	 */
/* cosQ = (eb,ea)	 */
/* sinQ = (eb,ebb)	 */
void compute_matrix_R(struct pe_matrix * pR, struct pe_vector * pvec_a,
		      struct pe_vector * pvec_bb, struct pe_vector * pvec_c,
		      struct pe_vector * pvec_b)
{
	/* sinQ */
	pR->a21 = pvec_b->x * pvec_bb->x + pvec_b->y * pvec_bb->y
		+ pvec_b->z * pvec_bb->z;
	/* cosQ */
	pR->a22 = pvec_b->x * pvec_a->x + pvec_b->y * pvec_a->y
		+ pvec_b->z * pvec_a->z;
	pR->a23 = 0.0;
	pR->a11 = pR->a22;
	pR->a12 = -1.0 * pR->a21;
	pR->a13 = 0.0;
	pR->a31 = 0.0;
	pR->a32 = 0.0;
	pR->a33 = 1.0;
	if (DEBUG) {
		/* sinQ*sinQ + cosQ*cosQ == 1 */
		if (FLOAT_NEQ((pR->a21 * pR->a21 + pR->a22 * pR->a22 - 1.0)
			      / my_coor.ave_dis)) {
			fprintf(STDERR,
				"compute_matrix_R() : ERROR   sinQ*sinQ + cosQ*cosQ != 1\n");
			exit(1);
		}
	}
	return;
}

/* MM(ea, ec, ebb) = (eb, ec, ec x eb) */
int check_rot_matrix(struct pe_matrix * pt_MM, struct pe_vector * pvec_a,
		     struct pe_vector * pvec_bb, struct pe_vector * pvec_c,
		     struct pe_vector * pvec_b)
{
	float wrk_x, wrk_y, wrk_z, wrk_length;
	float wrk2_x, wrk2_y, wrk2_z;
	
	if (!FLOAT_NEQ(my_coor.ave_dis)) {
		fprintf(STDERR, "check_rot_matrix() : ERROR  my_coor.ave_dis= %8.3f\n",
			my_coor.ave_dis);
		exit(1);
	}
	
	/* eb == MM(ea) */
	wrk_x = pt_MM->a11 * pvec_a->x + pt_MM->a12 * pvec_a->y
		+ pt_MM->a13 * pvec_a->z;
	wrk_y = pt_MM->a21 * pvec_a->x + pt_MM->a22 * pvec_a->y
		+ pt_MM->a23 * pvec_a->z;
	wrk_z = pt_MM->a31 * pvec_a->x + pt_MM->a32 * pvec_a->y
		+ pt_MM->a33 * pvec_a->z;
	wrk_length = (wrk_x - pvec_b->x) * (wrk_x - pvec_b->x)
		+ (wrk_y - pvec_b->y) * (wrk_y - pvec_b->y)
		+ (wrk_z - pvec_b->z) * (wrk_z - pvec_b->z);
	if (FLOAT_NEQ(wrk_length / my_coor.ave_dis)) {
		fprintf(STDERR, "check_rot_matrix() : ERROR   wrk_length = %f, my_coor.ave_dis = %f\n",
			wrk_length, my_coor.ave_dis);
		fprintf(STDERR, "check_rot_matrix() : ERROR   eb != MM(ea)\n");
		fprintf(STDERR, "check_rot_matrix() : ERROR   where   ea is\n");
		print_vec(pvec_a);
		fprintf(STDERR, "check_rot_matrix() : ERROR   and     eb is\n");
		print_vec(pvec_b);
		fprintf(STDERR,
			"check_rot_matrix() : ERROR   and     MM(ea) is\n");
		fprintf(STDOUT,
			"check_rot_matrix() :         vec = ( %8.3f, %8.3f, %8.3f )\n",
			wrk_x, wrk_y, wrk_z);
		fprintf(STDERR, "check_rot_matrix() : ERROR   and     MM is\n");
		print_mtrx(pt_MM);
		return FAIL;
	}
	/* ec == MM(ec) */
	wrk_x = pt_MM->a11 * pvec_c->x + pt_MM->a12 * pvec_c->y
		+ pt_MM->a13 * pvec_c->z;
	wrk_y = pt_MM->a21 * pvec_c->x + pt_MM->a22 * pvec_c->y
		+ pt_MM->a23 * pvec_c->z;
	wrk_z = pt_MM->a31 * pvec_c->x + pt_MM->a32 * pvec_c->y
		+ pt_MM->a33 * pvec_c->z;
	wrk_length = (wrk_x - pvec_c->x) * (wrk_x - pvec_c->x)
		+ (wrk_y - pvec_c->y) * (wrk_y - pvec_c->y)
		+ (wrk_z - pvec_c->z) * (wrk_z - pvec_c->z);
	if (FLOAT_NEQ(wrk_length / my_coor.ave_dis)) {
		fprintf(STDERR, "check_rot_matrix() : ERROR   ec != MM(ec)\n");
		fprintf(STDERR, "check_rot_matrix() : ERROR   where   ec is\n");
		print_vec(pvec_c);
		fprintf(STDERR,
			"check_rot_matrix() : ERROR   and     MM(ec) is\n");
		fprintf(STDOUT,
			"check_rot_matrix() :         vec = ( %8.3f, %8.3f, %8.3f )\n",
			wrk_x, wrk_y, wrk_z);
		fprintf(STDERR, "check_rot_matrix() : ERROR   and     MM is\n");
		print_mtrx(pt_MM);
		return FAIL;
	}
	/* ec x eb == MM(ebb) */
	wrk_x = pvec_c->y * pvec_b->z - pvec_c->z * pvec_b->y;
	wrk_y = pvec_c->z * pvec_b->x - pvec_c->x * pvec_b->z;
	wrk_z = pvec_c->x * pvec_b->y - pvec_c->y * pvec_b->x;
	wrk2_x = pt_MM->a11 * pvec_bb->x + pt_MM->a12 * pvec_bb->y
		+ pt_MM->a13 * pvec_bb->z;
	wrk2_y = pt_MM->a21 * pvec_bb->x + pt_MM->a22 * pvec_bb->y
		+ pt_MM->a23 * pvec_bb->z;
	wrk2_z = pt_MM->a31 * pvec_bb->x + pt_MM->a32 * pvec_bb->y
		+ pt_MM->a33 * pvec_bb->z;
	wrk_length = (wrk_x - wrk2_x) * (wrk_x - wrk2_x)
		+ (wrk_y - wrk2_y) * (wrk_y - wrk2_y)
		+ (wrk_z - wrk2_z) * (wrk_z - wrk2_z);
	if (FLOAT_NEQ(wrk_length / my_coor.ave_dis)) {
		fprintf(STDERR,
			"check_rot_matrix() : ERROR   ec x eb != MM(ebb)\n");
		fprintf(STDERR,
			"check_rot_matrix() : ERROR   where   ebb is\n");
		print_vec(pvec_bb);
		fprintf(STDERR,
			"check_rot_matrix() : ERROR   and     ec x eb is\n");
		fprintf(STDOUT,
			"check_rot_matrix() :         vec = ( %8.3f, %8.3f, %8.3f )\n",
			wrk_x, wrk_y, wrk_z);
		fprintf(STDERR,
			"check_rot_matrix() : ERROR   and     MM(eb) is\n");
		print_vec(pvec_a);
		fprintf(STDERR, "check_rot_matrix() : ERROR   and     MM is\n");
		print_mtrx(pt_MM);
		return FAIL;
	}
	return SUCCESS;
}

/* type = slt_tiles[t_ind].type */
/* MM(my_trn_table[type].d*) == slt_tile[t_ind].d* */
/* MM( my_trn_table[type].*** - my_trn_table[type].c*) + slt_tile[t_ind].c*
       == the position of a next tile */

/* MM(my_protein.atom[ind]) + tvec == slt_tiles[t_ind] */
/* my_protein.atom[ind] */
/* slt_tiles[t_ind] */ 
/* my_protein.atom[nxt_ind] */ 
void rotation_matrix(struct pe_matrix * pt_MM, int t_ind,
		     struct pe_slt_tile * slt_tiles)
{
	struct pe_vector vec_a, vec_b, vec_c, vec_bb;
	struct pe_matrix T, IT, R;
	struct pe_matrix M;
	int cur_type;
	float wrk_length;
	
	/* a = direction of the corresponding tile in the trn_table */
	cur_type = slt_tiles[t_ind].type;
	if (slt_tiles[t_ind].code == C_UP)
		set_vector(&vec_a, - my_trn_table[cur_type].bx,
			   - my_trn_table[cur_type].by,
			   - my_trn_table[cur_type].bz);
	else if (slt_tiles[t_ind].code == C_DOWN)
		set_vector(&vec_a, my_trn_table[cur_type].bx,
			   my_trn_table[cur_type].by,
			   my_trn_table[cur_type].bz);
	else {
		fprintf(STDERR,
			"rotation_matrix() : ERROR slt_tiles[%d].code == UNKOWN\n",
			 t_ind);
		exit(1);
	}
	/* b = direction of the current tile */
	set_vector(&vec_b, slt_tiles[t_ind].dx, slt_tiles[t_ind].dy,
		   slt_tiles[t_ind].dz);	
	/* c = a x b */
	set_vector(&vec_c, vec_a.y * vec_b.z - vec_a.z * vec_b.y,
		   vec_a.z * vec_b.x - vec_a.x * vec_b.z,
		   vec_a.x * vec_b.y - vec_a.y * vec_b.x);
	/* bb = c x a */
	set_vector(&vec_bb, vec_c.y * vec_a.z - vec_c.z * vec_a.y,
		   vec_c.z * vec_a.x - vec_c.x * vec_a.z,
		   vec_c.x * vec_a.y - vec_c.y * vec_a.x);
	if (0) {
		fprintf(STDERR, "rotation_matrix() : DEBUG vec_a\n");
		print_vec(&vec_a);
		fprintf(STDERR, "rotation_matrix() : DEBUG vec_b\n");
		print_vec(&vec_b);
		fprintf(STDERR, "rotation_matrix() : DEBUG vec_bb\n");
		print_vec(&vec_bb);
		fprintf(STDERR, "rotation_matrix() : DEBUG vec_c\n");
		print_vec(&vec_c);
	}
	/* T(vec_a, vec_bb, vec_c) = (ex, ey, ez) */
	compute_matrix_T(&T, &vec_a, &vec_bb, &vec_c);
	wrk_length = determinant(&T);
	if (FLOAT_NEQ(wrk_length / my_coor.ave_dis)) {
		inverse_mat(&IT, &T);
	} else {
		fprintf(STDERR, "rotation_matrix() : ERROR  |T|= %f <==> 0 \n",
			wrk_length);
		exit(1);
	}
	/* R = rotation around z-axis */
	compute_matrix_R(&R, &vec_a, &vec_bb, &vec_c, &vec_b);

	/* MM = IT*R*T */
	mat_mul(&M, &R, &T);		/* M = R*T */
	mat_mul(pt_MM, &IT, &M);	/* MM = IT*M */
	if (DEBUG) {
		if (check_rot_matrix(pt_MM, &vec_a, &vec_bb, &vec_c, &vec_b)
		    == FAIL)
			exit(1);
	}
	if (0) {
		fprintf(STDERR, "rotation_matrix() : Rotation matrix is\n");
		print_mtrx(pt_MM);
	}
	return;
}

/* rot_mat(my_trn_table[type].d*) == slt_tile[t_ind].d* */
void set_rotation_matrix(struct pe_slt_tile * slt_tiles)
{
	struct pe_matrix rot_mat;
	int t_ct, seq_ct, mod_ct;
	
	for (mod_ct = 1; mod_ct <= my_macro.mod_num; mod_ct++) {
		for (seq_ct = my_macro.model[mod_ct].seq_head;
		     seq_ct <= my_macro.model[mod_ct].seq_tail; seq_ct++) {
			if(my_macro.sequence[seq_ct].tile_head == 0)
				continue;	/* Skip empty sequence */
			for (t_ct = my_macro.sequence[seq_ct].tile_head;
			     t_ct <= my_macro.sequence[seq_ct].tile_tail;
			     t_ct++) {
				rotation_matrix(&rot_mat, t_ct, slt_tiles);
				slt_tiles[t_ct].rot_mat[0] = rot_mat.a11;
				slt_tiles[t_ct].rot_mat[1] = rot_mat.a21;
				slt_tiles[t_ct].rot_mat[2] = rot_mat.a31;
				slt_tiles[t_ct].rot_mat[3] = 0.0;
				slt_tiles[t_ct].rot_mat[4] = rot_mat.a12;
				slt_tiles[t_ct].rot_mat[5] = rot_mat.a22;
				slt_tiles[t_ct].rot_mat[6] = rot_mat.a32;
				slt_tiles[t_ct].rot_mat[7] = 0.0;
				slt_tiles[t_ct].rot_mat[8] = rot_mat.a13;
				slt_tiles[t_ct].rot_mat[9] = rot_mat.a23;
				slt_tiles[t_ct].rot_mat[10] = rot_mat.a33;
				slt_tiles[t_ct].rot_mat[11] = 0.0;
				slt_tiles[t_ct].rot_mat[12] = 0.0;
				slt_tiles[t_ct].rot_mat[13] = 0.0;
				slt_tiles[t_ct].rot_mat[14] = 0.0;
				slt_tiles[t_ct].rot_mat[15] = 1.0;
			}
		}
		break;
	}
	return;
}

